/*		query_string_routines.h	Sep 04				*/
/*	Watch this space!						*/
/* (w/my luck, after going to the trouble of modifying everything in	*/
/* anticipation of needing a .h, I'll never actually need this file )	*/
#define QUERY_STRINGH_VERSION "query_string_routines.h version 1.0  3 Sep 2004"
