/*  err_stub.c	Aug 2004.  WJS					*/
/*    Routine to stick in old programs that use lib subrs	*/
/*  which now require an err function to print their errors	*/
/*  This function assumes a non-web environment; messages go	*/
/*  to stdout, etc.  Old programs that are inners should use	*/
/*  err_stub_methods instead of this program			*/

#include <stdio.h>
#include <stdlib.h>

void err(s,t)
char *s,*t;
{
  printf ("%s\n%s\n",s,t);
  exit(1);
}
