/* 
 default.h   - to be #include'd with "inners".c and outer.c         
    Define minimum values, both for default and for sanity
    Values are those of 23 Jan 96 outerw (wjs)

 18 Feb 05.  TOKEN 240 -> 384; DATUMSIZE -> 385; OUTBUFSIZE 240 -> 386  RCG
 13 Jun 03.  TOKEN 120 -> 240; DATUMSIZE; OUTBUFSIZE 120 -> 240  RCG
 15 Nov 00.  ATTRSIZE boosted to accomodate merged object attributes CLH
  3 May 99.  Removed OBJECTPARAMSIZE - unused (as many others may be...) WJS
 30 Jul 96.  TOKEN; DATUMSIZE; OUTBUFSIZE -> 120.  WJS
*/

/* used to save HTML QUERY_STRING value */
#define PARSAVSIZE 1024

/* used for storing HTML PATH_INFO environment value */
#define PATHSIZE 1024

/* used to store the value of input strings between parameter separators 
     this could include an attribute string */
#define TOKEN 384

/* a datum, or value size */
#define DATUMSIZE 385

/* size that a parameter name can be */
#define VARNAMESIZE 40

/* attribute size:  for a single datum all attributes defined */
#define ATTRSIZE 80

/* size of all attributes for a single parameter, including punctuation */
#define TOTATTRSIZE 1024

/* comment lines:  for a single comment line, all chars, excl. #, incl nl */
#define COMMENTLINE 80

/* buffer used to output a datum or a comment in output routines */
#define OUTBUFSIZE 386

/* used to output a comment or an objectname plus its parameters */
#define OUTVARBUFSIZE 80

