/*	error_exit_defn.h  v 1.0  20 Apr 99	WJS			*/

/*	Define error exit status for program.  Problem from my per-	*/
/*  spective was that too many "things" used 1 (which, by the way, is	*/
/*  defined in stdlib.h as EXIT_FAILURE, so 1 is in fact standard in	*/
/*  some way).  I wanted a status that meant that the program had in-	*/
/*  deed begun and chosen to fail.  For example, if you try to exe-	*/
/*  cute a program in a child process and give the wrong file spec, you	*/
/*  will get a 1 back from whatever component tried to find the pro-	*/
/*  gram, which is too easily confused with the program exiting with	*/
/*  a 1.								*/
/*	Because the wait system service, which reports child process	*/
/*  termination information, only reports the low order 8 bits of exit	*/
/*  status, I wanted a number < 255.  0 is reserved for success.	*/
/*  250 chosen after looking through a couple of errno.h files for	*/
/*  unused statuses between 1 & 255.  No guarantee that it will stay	*/
/*  unused, of course.  255 ruled out because perl die function uses it	*/

#ifndef ERROR_EXIT_STATUS
#define ERROR_EXIT_STATUS 250
#endif

#if ERROR_EXIT_STATUS < 0
#error "ERROR_EXIT_STATUS < 0 will most likely be interpreted as pos #"
#endif
#if ERROR_EXIT_STATUS == 0
#error "ERROR_EXIT_STATUS = 0 conflicts with normal exit status"
#endif
#if ERROR_EXIT_STATUS > 255
#error "ERROR_EXIT_STATUS > 255 causes trouble when issued from child processes"
#endif
