% an example for simple plotting of data
clear all
loadjg('//usjgofs.whoi.edu/jgofs/eqpac/tt007/ammonium','NaN')
whos;

% remap event to yearday
day=eventtoday(event);

% display
figure(1)
plot(nh4,-depth);
xlabel('NH4');ylabel('Depth');
figure(2);
plot3(nh4,day,-depth);
view(30,30);
axis([0,1.2,35,70,-160,0]);
xlabel('NH4');ylabel('Day');zlabel('Depth');





