cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                     Version 1.1                        cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
   

        subroutine fixstr(s)
        character*(*) s
        l=len(s)
        i=lnblnk(s)
        if (i.lt.l) then 
          i=i+1
        else 
          i=l
        endif
        s(i:i)=char(0)
        return
        end

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

        
        subroutine unfixstr(s)
        character*(*) s
        integer i
        i=index(s,char(0))
        if (i.gt.0)then
           s(i:len(s))=' '
        endif
        return
        end

