/*  err_stub_methods.c	Aug 2004.  WJS				*/
/*    Routine to stick in old methods that use lib subrs	*/
/*  which now require an err function to print their errors	*/
/*  Methods presumably contain outer, which defines error_,	*/
/*  which takes care of output in a web environment		*/

void error_();

void err(s,t)
char *s,*t;
{
  error_(s,t);
  return;
}
