% LOADJG   Retrieve variables from data objects over the network.
%    LOADJG(fname)
%       If 'fname' ends in .mat, the operation is the
%          same as  LOAD  fname.
%       Otherwise, it is assumed to be a 
%          data object and is retrieved from the network. LOADJG creates
%          a variable for each variable in the object and fills it with
%          the values. The object is "flattened" out, so that all the
%          vectors are the full length.
%
%    LOADJG(fname,modifier)
%       The modifier string affects how the results are stored. It
%          contains comma-separated keywords:
%          'NaN' separates different groups with a row of NaN's
%             so that plots will break appropriately
%  perhaps to come...
%          'level' generates a set of variables 
%                 level0, level1,..., levelN with N=# levels -1
%             which contain indices to the positions in the data 
%             vector where a variable at the corresponding level
%             has changed. The innermost level changes at each
%             row and does not have a pointer.
%          'nd=...' specifies the transformation of nd (no data)
%             values. Default is NaN.
%          'break=variablename' converts each variable to a matrix
%             with the different columns corresponding to different
%             values of the chosen variablename. Columns will be
%             filled at the end with no data values.
