cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                     Version 1.1                        cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

        integer function lnblnk(msg)
        character*(*) msg
        do 5 i=len(msg),1,-1
        if (msg(i:i).ne.' ') goto 10
5       continue
10      lnblnk=i
        return
        end

        integer function fort_close()
        call exit(0)
        fort_close=0
        end
