#include <string.h>
int ioopen_(s,nparams,ntotal)
char *s[];
int *nparams,*ntotal;
{
char *par;
int i,j;
par=(char *)malloc(*nparams*80);
for (i=0;i< *nparams;i++)strcpy(par+80*i,s[i]);
j=ioopenf_(par,nparams,ntotal);
for(i=0;i< *nparams;i++)
  if(*(par+80*i) == 0)s[i][0]=0;
free(par);
return j;
}
