/*	jdbfuncdefns.h		Aug 2004.  WJS				*/
/*	  Exists mainly because of the underscores business - not sure	*/
/*	if it's worth it...						*/

#define JDBFUNCDEFNSH_VERSION "jdbfuncdefns.h  version 1.0  30 Aug 2004"

#include "core.h"
#define FULL_JDBFUNCDEFNSH_VERSION JDBFUNCDEFNSH_VERSION"/"COREH_VERSION

/*  Don't see why we should go through this .h file "more than once".	*/
/*  Accordingly, let's bomb if UNDERSCORES is define.  Then we'll see	*/
/*  why!								*/

/*  HP & IBM from core.h						*/
#if HP
#define UNDERSCORES 0
#elif IBM
#define UNDERSCORES 0
#else
#define UNDERSCORES 1
#endif


/*  Idea is that IBM & HP don't allow underscores in their function 	*/
/*  defns... at least that's what I remember.  Don't know if this stuff	*/
/*  has ever been put on HP or IBM platforms, or if rules are still the	*/
/*  same, or if dependence is on OS rather than hardware, etc		*/

#if UNDERSCORES

int jdbopen_();
int jdbreada_();
int jdbread_();
int jdbclose_();
int jdbcomments_();
int jdblevel_();
int jdbattributes_();

#else

int jdbopen();
int jdbreada();
int jdbread();
int jdbclose();
int jdbcomments();
int jdblevel();
int jdbattributes();
#define jdbopen_ jdbopen
#define jdbreada_ jdbreada
#define jdbclose_ jdbclose
#define jdblevel_ jdblevel
#define jdbattributes_ jdbattributes
#define jdbcomments_ jdbcomments

#endif
