/* 
 default.h   - to be #include'd with "inners".c and outer.c         
    Define minimum values, both for default and for sanity
    Values are those of 23 Jan 96 outerw (wjs)

*****  PLEASE DO NOT EDIT THIS FILE
                     Copy it to a new file, edit that, and 
                     include it in the makefile as in the
                     example of  "def"
*/

/* used to store the value of input strings between parameter separators 
     this could include an attribute string */
#define TOKEN 80

/* a datum, or value size */
#define DATUMSIZE 80

/* size that a parameter name can be */
#define VARNAMESIZE 40

/* attribute size:  for a single datum all attributes defined */
#define ATTRSIZE 80

/* size of all attributes for a single parameter, including punctuation */
#define TOTATTRSIZE 1024

/* comment lines:  for a single comment line, all chars, excl. #, incl nl */
#define COMMENTLINE 80
